package de.schloenvoigt.sparqltesttool.control;

/********************************************************************
	
This file is part of SPARQLTestTool - A Tool for testing SPARQL Queries

Copyright (C) 2010 Steffen Schlönvoigt <steffen@schloenvoigt.de>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
 	
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.
 	
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*********************************************************************/ 

import java.io.StringReader;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import com.hp.hpl.jena.query.Query;
import com.hp.hpl.jena.query.QueryException;
import com.hp.hpl.jena.query.QueryExecution;
import com.hp.hpl.jena.query.QueryExecutionFactory;
import com.hp.hpl.jena.query.QueryFactory;
import com.hp.hpl.jena.query.QuerySolution;
import com.hp.hpl.jena.query.ResultSet;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.RDFNode;

import de.schloenvoigt.sparqltesttool.model.SPARQLTestToolModel;

/**
 * Implements the logic for the SPARQL Test Tool 
 */
public class SPARQLTestToolController 
{
	
// ------------------- ATTRIBUTES -------------------------------
	/**
	 * The Model, to which we write our results
	 */
	private SPARQLTestToolModel model;

// ------------------- ACCESSORS -------------------------------

	/**
	 * Sets the Model, that is to be handled by this controller 
	 */
	public void setModel( SPARQLTestToolModel model )
	{
		this.model = model;
	}


// ------------------- CONSTRUCTOR -------------------------------
	/**
	 * CONSTRUCTOR
	 * @param model
	 */
	public SPARQLTestToolController( SPARQLTestToolModel model )
	{
		this.setModel( model );
	}

	
// ------------------- METHODS -----------------------------------	
	
	/**
	 * This method does the actual Query.
	 * @param data   RDF data in TURTLE format
	 * @param query  A SPARQL Query
	 */
	public void query( String data, String query )
	{
		model.setInputFieldTexts(data, query);
		Model rdfData = ModelFactory.createDefaultModel();
		HtmlOutputWriter output = new HtmlOutputWriter();		

		// First try to read in the data
		if( data.trim() != "" )
		{
			StringReader dataReader = new StringReader(data );
			try{
				rdfData.read( dataReader, "", "TURTLE" );	
			}
			catch( Exception x )
			{				
				output.writeError( "Error reading Data input!").br()
				      .writeMessage( x.getLocalizedMessage() ).commit();				
				return;
			}
			output.writeSuccess( "Data read successfully.").br().br();
		}
		else
		{				
			output.writeWarning( "Data is empty." ).br().commit();			
			return;
		}

		// Now try to query
		if( query.trim() != "" )
		{
			// Bring the Namespaces of the Data model over to SPARQL
			// This way, they have to be entered only once
			String sparqlPrefixes = "";
			for( Map.Entry<String, String> prefix : rdfData.getNsPrefixMap().entrySet() )
			{
				sparqlPrefixes += "PREFIX  "+prefix.getKey()+": <"+prefix.getValue()+">\n";
			}


			Query arqQuery;
			// Create a query object
			try
			{
				arqQuery = QueryFactory.create( sparqlPrefixes + query  );

			}
			catch( QueryException qx )
			{
				output.writeError( "Error parsing Query!" ).br().
				       writeMessage( qx.getLocalizedMessage() ).commit();				
				return;
			}
			// Create an Execution object
			QueryExecution qexec = QueryExecutionFactory.create( arqQuery, rdfData );
			try
			{
				output.writeMessage( "Query Result:" ).br();
				// Now do the query depending on the type
				if( arqQuery.getQueryType() == Query.QueryTypeSelect )
				{
					ResultSet results = qexec.execSelect();				
					output.writeTableHeader( arqQuery.getResultVars() );
					ArrayList<String> cells = new ArrayList<String>( );
					while( results.hasNext() )			
					{	
						cells.clear();
						QuerySolution soln = results.nextSolution();					
						for( String varName : arqQuery.getResultVars() )
						{						
							RDFNode item = soln.get( varName );
							if( item != null )cells.add( item.toString() );
							else cells.add( "" );
						}
						output.writeTableLine(cells);
					}
					output.closeTable();
				}
				else if( arqQuery.getQueryType() == Query.QueryTypeAsk )
				{
					boolean askResult = qexec.execAsk( );
					output.writeMessage( askResult?"Yes":"No" );
				}
				else if( arqQuery.getQueryType() == Query.QueryTypeConstruct )
				{
					Model resultModel = qexec.execConstruct();
					StringWriter writer = new StringWriter();
					resultModel.write(writer, "TURTLE");
					output.writeMessage( writer.toString() );
				}
				else if( arqQuery.getQueryType() == Query.QueryTypeDescribe )
				{
					Model resultModel = qexec.execDescribe();
					StringWriter writer = new StringWriter();
					resultModel.write(writer, "TURTLE");
					output.writeMessage( writer.toString() );
				}
				qexec.close();
			}
			catch( Exception x )
			{
				output.clear().writeError( "Error while executing Query!" ).br().writeMessage( x.getLocalizedMessage() ).commit();				
				return;
			}
		}
		output.commit();	

	}
	
// -------------------------------- UTILITIES -------------------------
	
	/**
	 * Helper class, to generate the HTML for the output
	 */
	private class HtmlOutputWriter
	{
		// The buffer
		private String output;
		/**
		 * CONSTRUCTOR
		 */
		public HtmlOutputWriter()
		{
			this.output = "";
		}

		/**
		 * Clears the Buffer
		 * @return this instance
		 */
		public HtmlOutputWriter clear( )
		{
			this.output = "";
			return this;
		}

		/**
		 * Write a normal message
		 * @param message
		 * @return this instance
		 */
		public HtmlOutputWriter writeMessage( String message )
		{
			output += escapeHtmlFull( message );
			return this;
		}

		/**
		 * Do a line break
		 * @return  this instance
		 */
		public HtmlOutputWriter br( )
		{
			output += "<br>\n";
			return this;
		}

		/**
		 * Write a success message (will be colored green)
		 * @param message
		 * @return this instance
		 */
		public HtmlOutputWriter writeSuccess( String message )
		{
			output += "<b color=green>"+escapeHtmlFull( message )+"</b>";
			return this;
		}

		/**
		 * Write an error message (will be colored red)
		 * @param message
		 * @return  this instance
		 */
		public HtmlOutputWriter writeError( String message )
		{
			output += "<b color=red>"+escapeHtmlFull( message )+"</b>";
			return this;
		}

		/**
		 * Write a warning message (will be colored yellow)
		 * @param message
		 * @return  this instance
		 */
		public HtmlOutputWriter writeWarning( String message )
		{
			output += "<b color=orange>"+escapeHtmlFull( message )+"</b>";
			return this;
		}

		/**
		 * Write the Head of a table
		 * @param colnames The headings that should appear
		 */
		public void writeTableHeader( List<String> colnames )
		{
			output += "<table border=1><tr>";
			for( String varName : colnames )
			{
				output +="<th>"+varName+"</th>";
			}
			output += "</tr>\n";
		}

		/**
		 * Write a row in the table
		 * @param cells contents of the row
		 */
		public void writeTableLine( List<String> cells )
		{
			output += "<tr>";
			for( String cell : cells )
			{
				output +="<td>"+cell+"</td>";
			}
			output += "</tr>\n";			
		}

		/**
		 * Finish the table
		 * @return  this instance
		 */
		public HtmlOutputWriter closeTable()
		{
			output += "</table>";
			return this;
		}

		/**
		 * Commit the changes (writes them to the model)
		 */
		public void commit()
		{
			model.setOutputString(output);
		}

		/**
		 * Returns the current content of the buffer 
		 */
		public String toString( ){ return this.output; }
		

		/**
		 * Removes unwanted characters from strings or
		 * encodes them in a HTML friendly way
		 * @param s A string that should be escaped
		 * @return  The escaped string
		 */
		public String escapeHtmlFull(String s)
		{
			StringBuilder sb = new StringBuilder(s.length());
			for (int i = 0; i < s.length(); i++)
			{
				char ch = s.charAt(i);
				if (ch >= 'a' && ch <= 'z' || ch >= 'A' && ch <= 'Z' || ch >= '0' && ch <= '9')
				{
					// safe
					sb.append(ch);
				}
				else if (ch == '\n')
				{
					// translate linebreaks to HTML
					sb.append("<br>\n");
				}
				else if (Character.isWhitespace(ch))
				{
					// paranoid version: whitespaces are unsafe - escape
					// conversion of (int)ch is naive
					sb.append("&#").append((int) ch).append(";");
				}
				else if (Character.isISOControl(ch))
				{
					// paranoid version:isISOControl which are not isWhitespace removed !
					// do nothing do not include in output !
				}
				else if (Character.isHighSurrogate(ch))
				{
					int codePoint;
					if (i + 1 < s.length() && Character.isSurrogatePair(ch, s.charAt(i + 1))
							&& Character.isDefined(codePoint = (Character.toCodePoint(ch, s.charAt(i + 1)))))
					{
						sb.append("&#").append(codePoint).append(";");
					}		         
					i++; //in sboth ways move forward
				}
				else if(Character.isLowSurrogate(ch))
				{		         
					i++; // move forward,do nothing do not include in output !
				}
				else
				{
					if (Character.isDefined(ch))
					{
						// paranoid version
						// the rest is unsafe, including <127 control chars
						sb.append("&#").append((int) ch).append(";");
					}
					//do nothing do not include undefined in output!
				}
			}
			return sb.toString();
		}


	}


}
